#!/usr/bin/python

import sys, re;
from structs import unions, structs, defines;

# command line arguments
arch    = sys.argv[1];
outfile = sys.argv[2];
infiles = sys.argv[3:];


###########################################################################
# configuration #2: architecture information

inttypes = {};
header = {};
footer = {};

#arm
inttypes["arm32"] = {
    "unsigned long" : "__danger_unsigned_long_on_arm32",
    "long"          : "__danger_long_on_arm32",
    "xen_pfn_t"     : "__align8__ uint64_t",
    "xen_ulong_t"   : "__align8__ uint64_t",
    "uint64_t"      : "__align8__ uint64_t",
};
header["arm32"] = """
#define __arm___ARM32 1
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(n64, n32) union { uint64_t n64; uint32_t n32; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(n64, n32) uint64_t n64
# define __align8__ FIXME
#endif
""";
footer["arm32"] = """
#undef __DECL_REG
"""

inttypes["arm64"] = {
    "unsigned long" : "__danger_unsigned_long_on_arm64",
    "long"          : "__danger_long_on_arm64",
    "xen_pfn_t"     : "__align8__ uint64_t",
    "xen_ulong_t"   : "__align8__ uint64_t",
    "uint64_t"      : "__align8__ uint64_t",
};
header["arm64"] = """
#define __aarch64___ARM64 1
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(n64, n32) union { uint64_t n64; uint32_t n32; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(n64, n32) uint64_t n64
# define __align8__ FIXME
#endif
""";
footer["arm64"] = """
#undef __DECL_REG
"""

# x86_32
inttypes["x86_32"] = {
    "unsigned long" : "uint32_t",
    "long"          : "uint32_t",
    "xen_pfn_t"     : "uint32_t",
    "xen_ulong_t"   : "uint32_t",
};
header["x86_32"] = """
#define __i386___X86_32 1
#pragma pack(4)
""";
footer["x86_32"] = """
#pragma pack()
""";

# x86_64
inttypes["x86_64"] = {
    "unsigned long" : "__align8__ uint64_t",
    "long"          : "__align8__ uint64_t",
    "xen_pfn_t"     : "__align8__ uint64_t",
    "xen_ulong_t"   : "__align8__ uint64_t",
};
header["x86_64"] = """
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
# define __DECL_REG(name) union { uint64_t r ## name, e ## name; }
# define __align8__ __attribute__((aligned (8)))
#else
# define __DECL_REG(name) uint64_t r ## name
# define __align8__ FIXME
#endif
#define __x86_64___X86_64 1
""";
footer["x86_64"] = """
#undef __DECL_REG
"""

###########################################################################
# main

input  = "";
output = "";
fileid = re.sub("[-.]", "_", "__FOREIGN_%s__" % outfile.upper());

# read input header files
for name in infiles:
    f = open(name, "r");
    input += f.read();
    f.close();

# add header
output += """
/*
 * public xen defines and struct for %s
 * generated by %s -- DO NOT EDIT
 */

#ifndef %s
#define %s 1

""" % (arch, sys.argv[0], fileid, fileid)

if arch in header:
    output += header[arch];
    output += "\n";

defined = {}

# add defines to output
for line in re.findall("#define[^\n]+", input):
    for define in defines:
        regex = "#define\s+%s\\b" % define;
        match = re.search(regex, line);
        if None == match:
            continue;
        defined[define] = 1
        if define.upper()[0] == define[0]:
            replace = define + "_" + arch.upper();
        else:
            replace = define + "_" + arch;
        regex = "\\b%s\\b" % define;
        output += re.sub(regex, replace, line) + "\n";
output += "\n";

# delete defines, comments, empty lines
input = re.sub("#define[^\n]+\n", "", input);
input = re.compile("/\*(.*?)\*/", re.S).sub("", input)
input = re.compile("\n\s*\n", re.S).sub("\n", input);

# add unions to output
for union in unions:
    regex = "union\s+%s\s*\{(.*?)\n\};" % union;
    match = re.search(regex, input, re.S)
    if None == match:
        output += "#define %s_has_no_%s 1\n" % (arch, union);
    else:
        output += "union %s_%s {%s\n};\n" % (union, arch, match.group(1));
    output += "\n";

# add structs to output
for struct in structs:
    regex = "(?:#ifdef ([A-Z_]+))?\nstruct\s+%s\s*\{(.*?)\n\};" % struct;
    match = re.search(regex, input, re.S)
    if None == match or \
           (match.group(1) is not None and match.group(1) not in defined):
        output += "#define %s_has_no_%s 1\n" % (arch, struct);
    else:
        output += "struct %s_%s {%s\n};\n" % (struct, arch, match.group(2));
        output += "typedef struct %s_%s %s_%s_t;\n" % (struct, arch, struct, arch);
    output += "\n";

# add footer
if arch in footer:
    output += footer[arch];
    output += "\n";
output += "#endif /* %s */\n" % fileid;

# replace: defines
for define in defines:
    if define.upper()[0] == define[0]:
        replace = define + "_" + arch.upper();
    else:
        replace = define + "_" + arch;
    output = re.sub("\\b%s\\b" % define, replace, output);

# replace: unions
for union in unions:
    output = re.sub("\\b(union\s+%s)\\b" % union, "\\1_%s" % arch, output);

# replace: structs + struct typedefs
for struct in structs:
    output = re.sub("\\b(struct\s+%s)\\b" % struct, "\\1_%s" % arch, output);
    output = re.sub("\\b(%s)_t\\b" % struct, "\\1_%s_t" % arch, output);

# replace: integer types
integers = inttypes[arch].keys();
integers.sort(lambda a, b: cmp(len(b),len(a)));
for type in integers:
    output = re.sub("\\b%s\\b" % type, inttypes[arch][type], output);

# print results
f = open(outfile, "w");
f.write(output);
f.close;




from distutils.core import setup, Extension
from distutils.ccompiler import new_compiler
import os
import sys

extra_compile_args  = [ "-fno-strict-aliasing", "-Werror" ]

XEN_ROOT = "../.."

fsimage = Extension("fsimage",
    extra_compile_args = extra_compile_args,
    include_dirs = [ XEN_ROOT + "/tools/libfsimage/common/" ],
    library_dirs = [ XEN_ROOT + "/tools/libfsimage/common/" ],
    libraries = ["fsimage"],
    sources = ["src/fsimage/fsimage.c"])

pkgs = [ 'grub' ]

setup(name='pygrub',
      version='0.3',
      description='Boot loader that looks a lot like grub for Xen',
      author='Jeremy Katz',
      author_email='katzj@redhat.com',
      license='GPL',
      package_dir={'grub': 'src', 'fsimage': 'src'},
      scripts = ["src/pygrub"],
      packages=pkgs,
      ext_modules = [ fsimage ]
      )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Unit tests for migration v2 streams
"""

import unittest

from struct import calcsize

from xen.migration import libxc, libxl

class TestLibxc(unittest.TestCase):

    def test_format_sizes(self):

        for fmt, sz in ( (libxc.IHDR_FORMAT, 24),
                         (libxc.DHDR_FORMAT, 16),
                         (libxc.RH_FORMAT, 8),

                         (libxc.PAGE_DATA_FORMAT, 8),
                         (libxc.X86_PV_INFO_FORMAT, 8),
                         (libxc.X86_PV_P2M_FRAMES_FORMAT, 8),
                         (libxc.X86_PV_VCPU_HDR_FORMAT, 8),
                         (libxc.TSC_INFO_FORMAT, 24),
                         (libxc.HVM_PARAMS_ENTRY_FORMAT, 16),
                         (libxc.HVM_PARAMS_FORMAT, 8),
                         ):
            self.assertEqual(calcsize(fmt), sz)


class TestLibxl(unittest.TestCase):

    def test_format_sizes(self):

        for fmt, sz in ( (libxl.HDR_FORMAT, 16),
                         (libxl.RH_FORMAT, 8),

                         (libxl.EMULATOR_HEADER_FORMAT, 8),
                         ):
            self.assertEqual(calcsize(fmt), sz)


def test_suite():
    suite = unittest.TestSuite()

    suite.addTest(unittest.makeSuite(TestLibxc))
    suite.addTest(unittest.makeSuite(TestLibxl))

    return suite

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Libxl Migration v2 streams

Record structures as per docs/specs/libxl-migration-stream.pandoc, and
verification routines.
"""

import sys

from struct import calcsize, unpack, unpack_from
from xen.migration.verify import StreamError, RecordError, VerifyBase
from xen.migration.libxc import VerifyLibxc

# Header
HDR_FORMAT = "!QII"

HDR_IDENT = 0x4c6962786c466d74 # "LibxlFmt" in ASCII
HDR_VERSION = 2

HDR_OPT_BIT_ENDIAN = 0
HDR_OPT_BIT_LEGACY = 1

HDR_OPT_LE     = (0 << HDR_OPT_BIT_ENDIAN)
HDR_OPT_BE     = (1 << HDR_OPT_BIT_ENDIAN)
HDR_OPT_LEGACY = (1 << HDR_OPT_BIT_LEGACY)

HDR_OPT_RESZ_MASK = 0xfffc

# Records
RH_FORMAT = "II"

REC_TYPE_end                    = 0x00000000
REC_TYPE_libxc_context          = 0x00000001
REC_TYPE_emulator_xenstore_data = 0x00000002
REC_TYPE_emulator_context       = 0x00000003
REC_TYPE_checkpoint_end         = 0x00000004

rec_type_to_str = {
    REC_TYPE_end                    : "End",
    REC_TYPE_libxc_context          : "Libxc context",
    REC_TYPE_emulator_xenstore_data : "Emulator xenstore data",
    REC_TYPE_emulator_context       : "Emulator context",
    REC_TYPE_checkpoint_end         : "Checkpoint end",
}

# emulator_* header
EMULATOR_HEADER_FORMAT = "II"

EMULATOR_ID_unknown       = 0x00000000
EMULATOR_ID_qemu_trad     = 0x00000001
EMULATOR_ID_qemu_upstream = 0x00000002

emulator_id_to_str = {
    EMULATOR_ID_unknown       : "Unknown",
    EMULATOR_ID_qemu_trad     : "Qemu Traditional",
    EMULATOR_ID_qemu_upstream : "Qemu Upstream",
}


#
# libxl format
#

LIBXL_QEMU_SIGNATURE = "DeviceModelRecord0002"
LIBXL_QEMU_RECORD_HDR = "=%dsI" % (len(LIBXL_QEMU_SIGNATURE), )

class VerifyLibxl(VerifyBase):
    """ Verify a Libxl v2 stream """

    def __init__(self, info, read):
        VerifyBase.__init__(self, info, read)


    def verify(self):
        """ Verity a libxl stream """

        self.verify_hdr()

        while self.verify_record() != REC_TYPE_end:
            pass


    def verify_hdr(self):
        """ Verify a Header """
        ident, version, options = self.unpack_exact(HDR_FORMAT)

        if ident != HDR_IDENT:
            raise StreamError("Bad image id: Expected 0x%x, got 0x%x"
                              % (HDR_IDENT, ident))

        if version != HDR_VERSION:
            raise StreamError("Unknown image version: Expected %d, got %d"
                              % (HDR_VERSION, version))

        if options & HDR_OPT_RESZ_MASK:
            raise StreamError("Reserved bits set in image options field: 0x%x"
                              % (options & HDR_OPT_RESZ_MASK))

        if ( (sys.byteorder == "little") and
             ((options & HDR_OPT_BIT_ENDIAN) != HDR_OPT_LE) ):
            raise StreamError(
                "Stream is not native endianess - unable to validate")

        endian = ["little", "big"][options & HDR_OPT_LE]

        if options & HDR_OPT_LEGACY:
            self.info("Libxl Header: %s endian, legacy converted" % (endian, ))
        else:
            self.info("Libxl Header: %s endian" % (endian, ))


    def verify_record(self):
        """ Verify an individual record """
        rtype, length = self.unpack_exact(RH_FORMAT)

        if rtype not in rec_type_to_str:
            raise StreamError("Unrecognised record type %x" % (rtype, ))

        self.info("Libxl Record: %s, length %d"
                  % (rec_type_to_str[rtype], length))

        contentsz = (length + 7) & ~7
        content = self.rdexact(contentsz)

        padding = content[length:]
        if padding != "\x00" * len(padding):
            raise StreamError("Padding containing non0 bytes found")

        if rtype not in record_verifiers:
            raise RuntimeError("No verification function for libxl record '%s'"
                               % rec_type_to_str[rtype])
        else:
            record_verifiers[rtype](self, content[:length])

        return rtype


    def verify_record_end(self, content):
        """ End record """

        if len(content) != 0:
            raise RecordError("End record with non-zero length")


    def verify_record_libxc_context(self, content):
        """ Libxc context record """

        if len(content) != 0:
            raise RecordError("Libxc context record with non-zero length")

        # Verify the libxc stream, as we can't seek forwards through it
        VerifyLibxc(self.info, self.read).verify()


    def verify_record_emulator_xenstore_data(self, content):
        """ Emulator Xenstore Data record """
        minsz = calcsize(EMULATOR_HEADER_FORMAT)

        if len(content) < minsz:
            raise RecordError("Length must be at least %d bytes, got %d"
                              % (minsz, len(content)))

        emu_id, emu_idx = unpack(EMULATOR_HEADER_FORMAT, content[:minsz])

        if emu_id not in emulator_id_to_str:
            raise RecordError("Unrecognised emulator id 0x%x" % (emu_id, ))

        self.info("Emulator Xenstore Data (%s, idx %d)"
                  % (emulator_id_to_str[emu_id], emu_idx))

        # Chop off the emulator header
        content = content[minsz:]

        if len(content):

            if content[-1] != '\x00':
                raise RecordError("Data not NUL terminated")

            # Split without the final NUL, to get an even number of parts
            parts = content[:-1].split("\x00")

            if (len(parts) % 2) != 0:
                raise RecordError("Expected an even number of strings, got %d"
                                  % (len(parts), ))

            for key, val in zip(parts[0::2], parts[1::2]):
                self.info("  '%s' = '%s'" % (key, val))


    def verify_record_emulator_context(self, content):
        """ Emulator Context record """
        minsz = calcsize(EMULATOR_HEADER_FORMAT)

        if len(content) < minsz:
            raise RecordError("Length must be at least %d bytes, got %d"
                              % (minsz, len(content)))

        emu_id, emu_idx = unpack(EMULATOR_HEADER_FORMAT, content[:minsz])

        if emu_id not in emulator_id_to_str:
            raise RecordError("Unrecognised emulator id 0x%x" % (emu_id, ))

        self.info("  Index %d, type %s" % (emu_idx, emulator_id_to_str[emu_id]))


    def verify_record_checkpoint_end(self, content):
        """ Checkpoint end record """

        if len(content) != 0:
            raise RecordError("Checkpoint end record with non-zero length")


record_verifiers = {
    REC_TYPE_end:
        VerifyLibxl.verify_record_end,
    REC_TYPE_libxc_context:
        VerifyLibxl.verify_record_libxc_context,
    REC_TYPE_emulator_xenstore_data:
        VerifyLibxl.verify_record_emulator_xenstore_data,
    REC_TYPE_emulator_context:
        VerifyLibxl.verify_record_emulator_context,
    REC_TYPE_checkpoint_end:
        VerifyLibxl.verify_record_checkpoint_end,
}




from distutils.core import setup, Extension
import os, sys

XEN_ROOT = "../.."

extra_compile_args  = [ "-fno-strict-aliasing", "-Werror" ]

PATH_XEN      = XEN_ROOT + "/tools/include"
PATH_LIBXC    = XEN_ROOT + "/tools/libxc"
PATH_LIBXL    = XEN_ROOT + "/tools/libxl"
PATH_XENSTORE = XEN_ROOT + "/tools/xenstore"

xc = Extension("xc",
               extra_compile_args = extra_compile_args,
               include_dirs       = [ PATH_XEN, PATH_LIBXC + "/include", "xen/lowlevel/xc" ],
               library_dirs       = [ PATH_LIBXC ],
               libraries          = [ "xenctrl", "xenguest" ],
               depends            = [ PATH_LIBXC + "/libxenctrl.so", PATH_LIBXC + "/libxenguest.so" ],
               sources            = [ "xen/lowlevel/xc/xc.c" ])

xs = Extension("xs",
               extra_compile_args = extra_compile_args,
               include_dirs       = [ PATH_XEN, PATH_XENSTORE + "/include", "xen/lowlevel/xs" ],
               library_dirs       = [ PATH_XENSTORE ],
               libraries          = [ "xenstore" ],
               depends            = [ PATH_XENSTORE + "/libxenstore.so" ],
               sources            = [ "xen/lowlevel/xs/xs.c" ])

xl = Extension("xl",
               extra_compile_args = extra_compile_args,
               include_dirs       = [ PATH_XEN, PATH_LIBXL, PATH_LIBXC + "/include", "xen/lowlevel/xl" ],
               library_dirs       = [ PATH_LIBXL ],
               libraries          = [ "xenlight" ],
               depends            = [ PATH_LIBXL + "/libxenlight.so" ],
               sources            = [ "xen/lowlevel/xl/xl.c", "xen/lowlevel/xl/_pyxl_types.c" ])

plat = os.uname()[0]
modules = [ xc, xs ]
#modules.extend([ xl ])

setup(name            = 'xen',
      version         = '3.0',
      description     = 'Xen',
      packages        = ['xen',
                         'xen.migration',
                         'xen.lowlevel',
                        ],
      ext_package = "xen.lowlevel",
      ext_modules = modules
      )



#!/usr/bin/python

import sys

def read_levels():
	f = open('../../../libxc/include/xentoollog.h', 'r')

	levels = []
	record = False
	for l in f.readlines():
		if 'XTL_NUM_LEVELS' in l:
			break
		if record == True:
			levels.append(l.split(',')[0].strip())
		if 'XTL_NONE' in l:
			record = True

	f.close()

	olevels = [level[4:].capitalize() for level in levels]

	return levels, olevels

# .ml

def gen_ml(olevels):
	s = ""

	s += "type level = \n"
	for level in olevels:
		s += '\t| %s\n' % level

	s += "\nlet level_to_string level =\n"
	s +=  "\tmatch level with\n"
	for level in olevels:
		s += '\t| %s -> "%s"\n' % (level, level)

	s += "\nlet level_to_prio level =\n"
	s += "\tmatch level with\n"
	for index,level in enumerate(olevels):
		s += '\t| %s -> %d\n' % (level, index)

	return s

# .mli

def gen_mli(olevels):
	s = ""

	s += "type level = \n"
	for level in olevels:
		s += '\t| %s\n' % level

	return s

# .c

def gen_c(level):
	s = ""

	s += "static value Val_level(xentoollog_level c_level)\n"
	s += "{\n"
	s += "\tswitch (c_level) {\n"
	s += "\tcase XTL_NONE: /* Not a real value */\n"
	s += '\t\tcaml_raise_sys_error(caml_copy_string("Val_level XTL_NONE"));\n'
	s += "\t\tbreak;\n"

	for index,level in enumerate(levels):
		s += "\tcase %s:\n\t\treturn Val_int(%d);\n" % (level, index)

	s += """\tcase XTL_NUM_LEVELS: /* Not a real value! */
	\t\tcaml_raise_sys_error(
	\t\t\tcaml_copy_string("Val_level XTL_NUM_LEVELS"));
	#if 0 /* Let the compiler catch this */
	\tdefault:
	\t\tcaml_raise_sys_error(caml_copy_string("Val_level Unknown"));
	\t\tbreak;
	#endif
	\t}
	\tabort();
	}
	"""

	return s

def autogen_header(open_comment, close_comment):
    s = open_comment + " AUTO-GENERATED FILE DO NOT EDIT " + close_comment + "\n"
    s += open_comment + " autogenerated by \n"
    s += reduce(lambda x,y: x + " ", range(len(open_comment + " ")), "")
    s += "%s" % " ".join(sys.argv)
    s += "\n " + close_comment + "\n\n"
    return s

if __name__ == '__main__':
	if len(sys.argv) < 3:
		print >>sys.stderr, "Usage: genlevels.py <mli> <ml> <c-inc>"
		sys.exit(1)

	levels, olevels = read_levels()

	_mli = sys.argv[1]
	mli = open(_mli, 'w')
	mli.write(autogen_header("(*", "*)"))

	_ml = sys.argv[2]
	ml = open(_ml, 'w')
	ml.write(autogen_header("(*", "*)"))

	_cinc = sys.argv[3]
	cinc = open(_cinc, 'w')
	cinc.write(autogen_header("/*", "*/"))

	mli.write(gen_mli(olevels))
	mli.write("\n")

	ml.write(gen_ml(olevels))
	ml.write("\n")

	cinc.write(gen_c(levels))
	cinc.write("\n")

	ml.write("(* END OF AUTO-GENERATED CODE *)\n")
	ml.close()
	mli.write("(* END OF AUTO-GENERATED CODE *)\n")
	mli.close()
	cinc.close()




#!/usr/bin/env python

import re,sys

pats = [
 [ r"^\s*#\s*include\s+", r"__InClUdE__ " ],
 [ r"^\s*#\s*ifdef (XEN_HAVE.*)\s+", r"__IfDeF__ \1" ],
 [ r"^\s*#\s*else /\* (XEN_HAVE.*) \*/\s+", r"__ElSe__" ],
 [ r"^\s*#\s*endif /\* (XEN_HAVE.*) \*/\s+", r"__EnDif__" ],
 [ r"^\s*#\s*define\s+([A-Z_]*_GUEST_HANDLE)", r"#define HIDE_\1" ],
 [ r"^\s*#\s*define\s+([a-z_]*_guest_handle)", r"#define hide_\1" ],
 [ r"XEN_GUEST_HANDLE(_[0-9A-Fa-f]+)?", r"COMPAT_HANDLE" ],
];

xlats = []

xlatf = open('xlat.lst', 'r')
for line in xlatf.readlines():
    match = re.subn(r"^\s*\?\s+(\w*)\s.*", r"\1", line.rstrip())
    if match[1]:
        xlats.append(match[0])
xlatf.close()

for line in sys.stdin.readlines():
    for pat in pats:
        line = re.subn(pat[0], pat[1], line)[0]
    for xlat in xlats:
        line = re.subn(r"(struct|union)\s+(%s|xen_%s)\s+(\w)" % (xlat, xlat),
            r"\1 @KeeP@\2 \3", line.rstrip())[0]
    print line.rstrip()



#!/usr/bin/env python
import sys

chars_per_line = 18
chars_so_far = 0

sys.stdout.write('"')

for char in sys.stdin.read():

    sys.stdout.write("\\%03o" % ord(char))
    chars_so_far = chars_so_far + 1

    if chars_so_far == chars_per_line:
        chars_so_far = 0
        sys.stdout.write('" \\\n"')

sys.stdout.write('"\n')



#============================================================================
# This library is free software; you can redistribute it and/or
# modify it under the terms of version 2.1 of the GNU Lesser General Public
# License as published by the Free Software Foundation.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; If not, see <http://www.gnu.org/licenses/>.
#============================================================================
# Copyright (C) 2009 flonatel GmbH & Co. KG
#============================================================================

import unittest

suite = unittest.TestSuite([])

if __name__ == "__main__":
    testresult = unittest.TextTestRunner(verbosity=3).run(suite)




#!/usr/bin/python

import sys,os

import idl

# typename -> ( ocaml_type, c_from_ocaml, ocaml_from_c )
builtins = {
    "bool":                 ("bool",                   "%(c)s = Bool_val(%(o)s)",           "Val_bool(%(c)s)" ),
    "int":                  ("int",                    "%(c)s = Int_val(%(o)s)",            "Val_int(%(c)s)"  ),
    "char *":               ("string option",          "%(c)s = String_option_val(%(o)s)",  "Val_string_option(%(c)s)"),
    "libxl_domid":          ("domid",                  "%(c)s = Int_val(%(o)s)",            "Val_int(%(c)s)"  ),
    "libxl_devid":          ("devid",                  "%(c)s = Int_val(%(o)s)",            "Val_int(%(c)s)"  ),
    "libxl_defbool":        ("bool option",            "%(c)s = Defbool_val(%(o)s)",        "Val_defbool(%(c)s)" ),
    "libxl_uuid":           ("int array",              "Uuid_val(&%(c)s, %(o)s)",   "Val_uuid(&%(c)s)"),
    "libxl_bitmap":         ("bool array",             "Bitmap_val(ctx, &%(c)s, %(o)s)",   "Val_bitmap(&%(c)s)"),    
    "libxl_key_value_list": ("(string * string) list", "libxl_key_value_list_val(&%(c)s, %(o)s)", "Val_key_value_list(&%(c)s)"),
    "libxl_string_list":    ("string list",            "libxl_string_list_val(&%(c)s, %(o)s)", "Val_string_list(&%(c)s)"),
    "libxl_mac":            ("int array",              "Mac_val(&%(c)s, %(o)s)",    "Val_mac(&%(c)s)"),
    "libxl_hwcap":          ("int32 array",            None,                                "Val_hwcap(&%(c)s)"),
    "libxl_ms_vm_genid":    ("int array",              "Ms_vm_genid_val(&%(c)s, %(o)s)",    "Val_ms_vm_genid(&%(c)s)"),
    # The following needs to be sorted out later
    "libxl_cpuid_policy_list": ("unit",                "%(c)s = 0",                         "Val_unit"),
    }

DEVICE_FUNCTIONS = [ ("add",            ["ctx", "t", "domid", "?async:'a", "unit", "unit"]),
                     ("remove",         ["ctx", "t", "domid", "?async:'a", "unit", "unit"]),
                     ("destroy",        ["ctx", "t", "domid", "?async:'a", "unit", "unit"]),
                   ]
DEVICE_LIST =      [ ("list",           ["ctx", "domid", "t list"]),
                   ]

functions = { # ( name , [type1,type2,....] )
    "device_vfb":     DEVICE_FUNCTIONS,
    "device_vkb":     DEVICE_FUNCTIONS,
    "device_disk":    DEVICE_FUNCTIONS + DEVICE_LIST +
                      [ ("insert",         ["ctx", "t", "domid", "?async:'a", "unit", "unit"]),
                        ("of_vdev",        ["ctx", "domid", "string", "t"]),
                      ],
    "device_nic":     DEVICE_FUNCTIONS + DEVICE_LIST +
                      [ ("of_devid",       ["ctx", "domid", "int", "t"]),
                      ],
    "device_pci":     DEVICE_FUNCTIONS + DEVICE_LIST +
                      [ ("assignable_add",    ["ctx", "t", "bool", "unit"]),
                        ("assignable_remove", ["ctx", "t", "bool", "unit"]),
                        ("assignable_list",   ["ctx", "t list"]),
                      ],
    "dominfo":        [ ("list",           ["ctx", "t list"]),
                        ("get",            ["ctx", "domid", "t"]),
                      ],
    "physinfo":       [ ("get",            ["ctx", "t"]),
                      ],
    "cputopology":    [ ("get",            ["ctx", "t array"]),
                      ],
    "domain_sched_params":
                      [ ("get",            ["ctx", "domid", "t"]),
                        ("set",            ["ctx", "domid", "t", "unit"]),
                      ],
}
def stub_fn_name(ty, name):
    return "stub_xl_%s_%s" % (ty.rawname,name)
    
def ocaml_type_of(ty):
    if ty.rawname in ["domid","devid"]:
        return ty.rawname
    elif isinstance(ty,idl.UInt):
        if ty.width in [8, 16]:
            # handle as ints
            width = None
        elif ty.width in [32, 64]:
            width = ty.width
        else:
            raise NotImplementedError("Cannot handle %d-bit int" % ty.width)
        if width:
            return "int%d" % ty.width
        else:
            return "int"
    elif isinstance(ty,idl.Array):
        return "%s array" % ocaml_type_of(ty.elem_type)
    elif isinstance(ty,idl.Builtin):
        if not builtins.has_key(ty.typename):
            raise NotImplementedError("Unknown Builtin %s (%s)" % (ty.typename, type(ty)))
        typename,_,_ = builtins[ty.typename]
        if not typename:
            raise NotImplementedError("No typename for Builtin %s (%s)" % (ty.typename, type(ty)))
        return typename
    elif isinstance(ty,idl.KeyedUnion):
        return ty.union_name
    elif isinstance(ty,idl.Aggregate):
        if ty.rawname is None:
            return ty.anon_struct
        else:
            return ty.rawname.capitalize() + ".t"
    else:
        return ty.rawname

ocaml_keywords = ['and', 'as', 'assert', 'begin', 'end', 'class', 'constraint',
    'do', 'done', 'downto', 'else', 'if', 'end', 'exception', 'external', 'false',
    'for', 'fun', 'function', 'functor', 'if', 'in', 'include', 'inherit',
    'initializer', 'lazy', 'let', 'match', 'method', 'module', 'mutable', 'new',
    'object', 'of', 'open', 'or', 'private', 'rec', 'sig', 'struct', 'then', 'to',
    'true', 'try', 'type', 'val', 'virtual', 'when', 'while', 'with']

def munge_name(name):
    if name in ocaml_keywords:
        return "xl_" + name
    else:
        return name

def ocaml_instance_of_field(f):
    if isinstance(f.type, idl.KeyedUnion):
        name = f.type.keyvar.name
    else:
        name = f.name
    return "%s : %s" % (munge_name(name), ocaml_type_of(f.type))

def gen_struct(ty, indent):
    s = ""
    for f in ty.fields:
        if f.type.private:
            continue
        x = ocaml_instance_of_field(f)
        x = x.replace("\n", "\n"+indent)
        s += indent + x + ";\n"
    return s

def gen_ocaml_keyedunions(ty, interface, indent, parent = None):
    s = ""
    union_type = ""
    
    if ty.rawname is not None:
        # Non-anonymous types need no special handling
        pass
    elif isinstance(ty, idl.KeyedUnion):
        if parent is None:
            nparent = ty.keyvar.name
        else:
            nparent = parent + "_" + ty.keyvar.name

        for f in ty.fields:
            if f.type is None: continue
            if f.type.rawname is not None: continue
            if isinstance(f.type, idl.Struct) and not f.type.has_fields(): continue
            s += "\ntype %s_%s =\n" % (nparent,f.name)
            s += "{\n"
            s += gen_struct(f.type, indent + "\t")
            s += "}\n"

        name = "%s__union" % ty.keyvar.name
        s += "\n"
        s += "type %s = " % name
        u = []
        for f in ty.fields:
            if f.type is None:
                u.append("%s" % (f.name.capitalize()))
            elif isinstance(f.type, idl.Struct):
                if f.type.rawname is not None:
                    u.append("%s of %s" % (f.name.capitalize(), f.type.rawname.capitalize()))
                elif f.type.has_fields():
                    u.append("%s of %s_%s" % (f.name.capitalize(), nparent, f.name))
                else:
                    u.append("%s" % (f.name.capitalize()))
            else:
                raise NotImplementedError("Cannot handle KeyedUnion fields which are not Structs")
            
        s += " | ".join(u) + "\n"
        ty.union_name = name

        union_type = "?%s:%s" % (munge_name(nparent), ty.keyvar.type.rawname)

    if s == "":
        return None, None
    return s.replace("\n", "\n%s" % indent), union_type

def gen_ocaml_anonstruct(ty, interface, indent, parent = None):
    s= ""

    if ty.rawname is not None:
        # Non-anonymous types need no special handling
        pass
    elif isinstance(ty, idl.Struct):
        name = "%s__anon" % parent
        s += "type %s = {\n" % name
        s += gen_struct(ty, indent)
        s += "}\n"
        ty.anon_struct = name
    if s == "":
        return None
    s = indent + s
    return s.replace("\n", "\n%s" % indent)

def gen_ocaml_ml(ty, interface, indent=""):

    if interface:
        s = ("""(* %s interface *)\n""" % ty.typename)
    else:
        s = ("""(* %s implementation *)\n""" % ty.typename)
        
    if isinstance(ty, idl.Enumeration):
        s += "type %s = \n" % ty.rawname
        for v in ty.values:
            s += "\t | %s\n" % v.rawname

        if interface:
            s += "\nval string_of_%s : %s -> string\n" % (ty.rawname, ty.rawname)
        else:
            s += "\nlet string_of_%s = function\n" % ty.rawname
            for v in ty.values:
                s += '\t| %s -> "%s"\n' % (v.rawname, v.valuename)

    elif isinstance(ty, idl.Aggregate):
        s += ""
        
        if ty.typename is None:
            raise NotImplementedError("%s has no typename" % type(ty))
        else:

            module_name = ty.rawname[0].upper() + ty.rawname[1:]

            if interface:
                s += "module %s : sig\n" % module_name
            else:
                s += "module %s = struct\n" % module_name
                
        # Handle KeyedUnions...
        union_types = []
        for f in ty.fields:
            ku, union_type = gen_ocaml_keyedunions(f.type, interface, "\t")
            if ku is not None:
                s += ku
                s += "\n"
            if union_type is not None:
                union_types.append(union_type)

        # Handle anonymous structs...
        for f in ty.fields:
            anon = gen_ocaml_anonstruct(f.type, interface, "\t", f.name)
            if anon is not None:
                s += anon
                s += "\n"

        s += "\ttype t =\n"
        s += "\t{\n"
        s += gen_struct(ty, "\t\t")
        s += "\t}\n"

        if ty.init_fn is not None:
            union_args = "".join([u + " -> " for u in union_types])
            if interface:
                s += "\tval default : ctx -> %sunit -> t\n" % union_args
            else:
                s += "\texternal default : ctx -> %sunit -> t = \"stub_libxl_%s_init\"\n" % (union_args, ty.rawname)

        if functions.has_key(ty.rawname):
            for name,args in functions[ty.rawname]:
                s += "\texternal %s : " % name
                s += " -> ".join(args)
                s += " = \"%s\"\n" % stub_fn_name(ty,name)
        
        s += "end\n"

    else:
        raise NotImplementedError("%s" % type(ty))
    return s.replace("\n", "\n%s" % indent)

def c_val(ty, c, o, indent="", parent = None):
    s = indent
    if isinstance(ty,idl.UInt):
        if ty.width in [8, 16]:
            # handle as ints
            width = None
        elif ty.width in [32, 64]:
            width = ty.width
        else:
            raise NotImplementedError("Cannot handle %d-bit int" % ty.width)
        if width:
            s += "%s = Int%d_val(%s);" % (c, width, o)
        else:
            s += "%s = Int_val(%s);" % (c, o)
    elif isinstance(ty,idl.Builtin):
        if not builtins.has_key(ty.typename):
            raise NotImplementedError("Unknown Builtin %s (%s)" % (ty.typename, type(ty)))
        _,fn,_ = builtins[ty.typename]
        if not fn:
            raise NotImplementedError("No c_val fn for Builtin %s (%s)" % (ty.typename, type(ty)))
        s += "%s;" % (fn % { "o": o, "c": c })
    elif isinstance (ty,idl.Array):
        s += "{\n"
        s += "\tint i;\n"
        s += "\t%s = Wosize_val(%s);\n" % (parent + ty.lenvar.name, o)
        s += "\t%s = (%s) calloc(%s, sizeof(*%s));\n" % (c, ty.typename, parent + ty.lenvar.name, c)
        s += "\tfor(i=0; i<%s; i++) {\n" % (parent + ty.lenvar.name)
        s += c_val(ty.elem_type, c+"[i]", "Field(%s, i)" % o, indent="\t\t", parent=parent) + "\n"
        s += "\t}\n"
        s += "}\n"
    elif isinstance(ty,idl.Enumeration) and (parent is None):
        n = 0
        s += "switch(Int_val(%s)) {\n" % o
        for e in ty.values:
            s += "    case %d: *%s = %s; break;\n" % (n, c, e.name)
            n += 1
        s += "    default: failwith_xl(ERROR_FAIL, \"cannot convert value to %s\"); break;\n" % ty.typename
        s += "}"
    elif isinstance(ty, idl.KeyedUnion):
        s += "{\n"
        s += "\tif(Is_long(%s)) {\n" % o
        n = 0
        s += "\t\tswitch(Int_val(%s)) {\n" % o
        for f in ty.fields:
            if f.type is None or not f.type.has_fields():
                s += "\t\t    case %d: %s = %s; break;\n" % (n,
                                                    parent + ty.keyvar.name,
                                                    f.enumname)
                n += 1
        s += "\t\t    default: failwith_xl(ERROR_FAIL, \"variant handling bug %s%s (long)\"); break;\n" % (parent, ty.keyvar.name)        
        s += "\t\t}\n"
        s += "\t} else {\n"
        s += "\t\t/* Is block... */\n"
        s += "\t\tswitch(Tag_val(%s)) {\n" % o
        n = 0
        for f in ty.fields:
            if f.type is not None and f.type.has_fields():
                if f.type.private:
                    continue
                s += "\t\t    case %d:\n" % (n)
                s += "\t\t        %s = %s;\n" % (parent + ty.keyvar.name, f.enumname)
                (nparent,fexpr) = ty.member(c, f, False)
                s += "%s" % c_val(f.type, fexpr, "Field(%s, 0)" % o, indent=indent+"\t\t        ")
                s += "break;\n"
                n += 1
        s += "\t\t    default: failwith_xl(ERROR_FAIL, \"variant handling bug %s%s (block)\"); break;\n" % (parent, ty.keyvar.name)
        s += "\t\t}\n"
        s += "\t}\n"
        s += "}"
    elif isinstance(ty, idl.Aggregate) and (parent is None or ty.rawname is None):
        n = 0
        for f in ty.fields:
            if f.type.private:
                continue
            (nparent,fexpr) = ty.member(c, f, ty.rawname is not None)
            s += "%s\n" % c_val(f.type, fexpr, "Field(%s, %d)" % (o,n), parent=nparent)
            n = n + 1
    else:
        s += "%s_val(ctx, %s, %s);" % (ty.rawname, ty.pass_arg(c, parent is None, passby=idl.PASS_BY_REFERENCE), o)
    
    return s.replace("\n", "\n%s" % indent)

def gen_c_val(ty, indent=""):
    s = "/* Convert caml value to %s */\n" % ty.rawname
    
    s += "static int %s_val (libxl_ctx *ctx, %s, value v)\n" % (ty.rawname, ty.make_arg("c_val", passby=idl.PASS_BY_REFERENCE))
    s += "{\n"
    s += "\tCAMLparam1(v);\n"
    s += "\n"

    s += c_val(ty, "c_val", "v", indent="\t") + "\n"
    
    s += "\tCAMLreturn(0);\n"
    s += "}\n"
    
    return s.replace("\n", "\n%s" % indent)
    
def ocaml_Val(ty, o, c, indent="", parent = None):
    s = indent
    if isinstance(ty,idl.UInt):
        if ty.width in [8, 16]:
            # handle as ints
            width = None
        elif ty.width in [32, 64]:
            width = ty.width
        else:
            raise NotImplementedError("Cannot handle %d-bit int" % ty.width)
        if width:
            s += "%s = caml_copy_int%d(%s);" % (o, width, c)
        else:
            s += "%s = Val_int(%s);" % (o, c)
    elif isinstance(ty,idl.Builtin):
        if not builtins.has_key(ty.typename):
            raise NotImplementedError("Unknown Builtin %s (%s)" % (ty.typename, type(ty)))
        _,_,fn = builtins[ty.typename]
        if not fn:
            raise NotImplementedError("No ocaml Val fn for Builtin %s (%s)" % (ty.typename, type(ty)))
        s += "%s = %s;" % (o, fn % { "c": c })
    elif isinstance(ty, idl.Array):
        s += "{\n"
        s += "\t    int i;\n"
        s += "\t    CAMLlocal1(array_elem);\n"
        s += "\t    %s = caml_alloc(%s,0);\n" % (o, parent + ty.lenvar.name)
        s += "\t    for(i=0; i<%s; i++) {\n" % (parent + ty.lenvar.name)
        s += "\t        %s\n" % ocaml_Val(ty.elem_type, "array_elem", c + "[i]", "", parent=parent)
        s += "\t        Store_field(%s, i, array_elem);\n" % o
        s += "\t    }\n"
        s += "\t}"
    elif isinstance(ty,idl.Enumeration) and (parent is None):
        n = 0
        s += "switch(%s) {\n" % c
        for e in ty.values:
            s += "    case %s: %s = Val_int(%d); break;\n" % (e.name, o, n)
            n += 1
        s += "    default: failwith_xl(ERROR_FAIL, \"cannot convert value from %s\"); break;\n" % ty.typename
        s += "}"
    elif isinstance(ty, idl.KeyedUnion):
        n = 0
        m = 0
        s += "switch(%s) {\n" % (parent + ty.keyvar.name)
        for f in ty.fields:
            s += "\t    case %s:\n" % f.enumname
            if f.type is None:
                s += "\t        /* %d: None */\n" % n
                s += "\t        %s = Val_long(%d);\n" % (o,n)
                n += 1
            elif not f.type.has_fields():
                s += "\t        /* %d: Long */\n" % n
                s += "\t        %s = Val_long(%d);\n" % (o,n)
                n += 1
            else:
                s += "\t        /* %d: Block */\n" % m
                (nparent,fexpr) = ty.member(c, f, parent is None)
                s += "\t        {\n"
                s += "\t\t        CAMLlocal1(tmp);\n"
                s += "\t\t        %s = caml_alloc(%d,%d);\n" % (o, 1, m)
                s += ocaml_Val(f.type, 'tmp', fexpr, indent="\t\t        ", parent=nparent)
                s += "\n"
                s += "\t\t        Store_field(%s, 0, tmp);\n" % o
                s += "\t        }\n"
                m += 1
                #s += "\t        %s = caml_alloc(%d,%d);\n" % (o,len(f.type.fields),n)
            s += "\t        break;\n"
        s += "\t    default: failwith_xl(ERROR_FAIL, \"cannot convert value from %s\"); break;\n" % ty.typename
        s += "\t}"
    elif isinstance(ty,idl.Aggregate) and (parent is None or ty.rawname is None):
        s += "{\n"
        if ty.rawname is None:
            fn = "anon_field"
        else:
            fn = "%s_field" % ty.rawname
        s += "\tCAMLlocal1(%s);\n" % fn
        s += "\n"
        s += "\t%s = caml_alloc_tuple(%d);\n" % (o, len(ty.fields))
        
        n = 0
        for f in ty.fields:
            if f.type.private:
                continue

            (nparent,fexpr) = ty.member(c, f, parent is None)

            s += "\n"
            s += "\t%s\n" % ocaml_Val(f.type, fn, ty.pass_arg(fexpr, c), parent=nparent)
            s += "\tStore_field(%s, %d, %s);\n" % (o, n, fn)
            n = n + 1
        s += "}"
    else:
        s += "%s = Val_%s(%s);" % (o, ty.rawname, ty.pass_arg(c, parent is None))
    
    return s.replace("\n", "\n%s" % indent).rstrip(indent)

def gen_Val_ocaml(ty, indent=""):
    s = "/* Convert %s to a caml value */\n" % ty.rawname

    s += "static value Val_%s (%s)\n" % (ty.rawname, ty.make_arg(ty.rawname+"_c"))
    s += "{\n"
    s += "\tCAMLparam0();\n"
    s += "\tCAMLlocal1(%s_ocaml);\n" % ty.rawname

    s += ocaml_Val(ty, "%s_ocaml" % ty.rawname, "%s_c" % ty.rawname, indent="\t") + "\n"
    
    s += "\tCAMLreturn(%s_ocaml);\n" % ty.rawname
    s += "}\n"
    return s.replace("\n", "\n%s" % indent)

def gen_c_stub_prototype(ty, fns):
    s = "/* Stubs for %s */\n" % ty.rawname
    for name,args in fns:        
        # For N args we return one value and take N-1 values as parameters
        s += "value %s(" % stub_fn_name(ty, name)
        s += ", ".join(["value v%d" % v for v in range(1,len(args))])
        s += ");\n"
    return s

def gen_c_default(ty):
    s = "/* Get the defaults for %s */\n" % ty.rawname
    # Handle KeyedUnions...
    union_types = []
    for f in ty.fields:
        if isinstance(f.type, idl.KeyedUnion):
            union_types.append(f.type.keyvar)

    s += "value stub_libxl_%s_init(value ctx, %svalue unit)\n" % (ty.rawname,
        "".join(["value " + u.name + ", " for u in union_types]))
    s += "{\n"
    s += "\tCAMLparam%d(ctx, %sunit);\n" % (len(union_types) + 2, "".join([u.name + ", " for u in union_types]))
    s += "\tCAMLlocal1(val);\n"
    s += "\tlibxl_%s c_val;\n" % ty.rawname
    s += "\tlibxl_%s_init(&c_val);\n" % ty.rawname
    for u in union_types:
        s += "\tif (%s != Val_none) {\n" % u.name
        s += "\t\t%s c = 0;\n" % u.type.typename
        s += "\t\t%s_val(CTX, &c, Some_val(%s));\n" % (u.type.rawname, u.name)
        s += "\t\tlibxl_%s_init_%s(&c_val, c);\n" % (ty.rawname, u.name)
        s += "\t}\n"
    s += "\tval = Val_%s(&c_val);\n" % ty.rawname
    if ty.dispose_fn:
        s += "\tlibxl_%s_dispose(&c_val);\n" % ty.rawname
    s += "\tCAMLreturn(val);\n"
    s += "}\n"
    return s

def gen_c_defaults(ty):
    s = gen_c_default(ty)
    return s

def autogen_header(open_comment, close_comment):
    s = open_comment + " AUTO-GENERATED FILE DO NOT EDIT " + close_comment + "\n"
    s += open_comment + " autogenerated by \n"
    s += reduce(lambda x,y: x + " ", range(len(open_comment + " ")), "")
    s += "%s" % " ".join(sys.argv)
    s += "\n " + close_comment + "\n\n"
    return s

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print >>sys.stderr, "Usage: genwrap.py <idl> <mli> <ml> <c-inc>"
        sys.exit(1)

    (_,types) = idl.parse(sys.argv[1])

    # Do not generate these yet.
    blacklist = [
        "cpupoolinfo",
        "vcpuinfo",
        ]

    for t in blacklist:
        if t not in [ty.rawname for ty in types]:
            print "unknown type %s in blacklist" % t

    types = [ty for ty in types if not ty.rawname in blacklist]
    
    _ml = sys.argv[3]
    ml = open(_ml, 'w')
    ml.write(autogen_header("(*", "*)"))

    _mli = sys.argv[2]
    mli = open(_mli, 'w')
    mli.write(autogen_header("(*", "*)"))
    
    _cinc = sys.argv[4]
    cinc = open(_cinc, 'w')
    cinc.write(autogen_header("/*", "*/"))

    for ty in types:
        if ty.private:
            continue
        #sys.stdout.write(" TYPE    %-20s " % ty.rawname)
        ml.write(gen_ocaml_ml(ty, False))
        ml.write("\n")

        mli.write(gen_ocaml_ml(ty, True))
        mli.write("\n")
        
        if ty.marshal_in():
            cinc.write(gen_c_val(ty))
            cinc.write("\n")
        cinc.write(gen_Val_ocaml(ty))
        cinc.write("\n")
        if functions.has_key(ty.rawname):
            cinc.write(gen_c_stub_prototype(ty, functions[ty.rawname]))
            cinc.write("\n")
        if ty.init_fn is not None:
            cinc.write(gen_c_defaults(ty))
            cinc.write("\n")
        #sys.stdout.write("\n")
    
    ml.write("(* END OF AUTO-GENERATED CODE *)\n")
    ml.close()
    mli.write("(* END OF AUTO-GENERATED CODE *)\n")
    mli.close()
    cinc.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Libxc Migration v2 streams

Record structures as per docs/specs/libxc-migration-stream.pandoc, and
verification routines.
"""

import sys

from struct import calcsize, unpack

from xen.migration.verify import StreamError, RecordError, VerifyBase

# Image Header
IHDR_FORMAT = "!QIIHHI"

IHDR_MARKER  = 0xffffffffffffffff
IHDR_IDENT   = 0x58454E46 # "XENF" in ASCII
IHDR_VERSION = 2

IHDR_OPT_BIT_ENDIAN = 0
IHDR_OPT_LE = (0 << IHDR_OPT_BIT_ENDIAN)
IHDR_OPT_BE = (1 << IHDR_OPT_BIT_ENDIAN)

IHDR_OPT_RESZ_MASK = 0xfffe

# Domain Header
DHDR_FORMAT = "IHHII"

DHDR_TYPE_x86_pv  = 0x00000001
DHDR_TYPE_x86_hvm = 0x00000002
DHDR_TYPE_x86_pvh = 0x00000003
DHDR_TYPE_arm     = 0x00000004

dhdr_type_to_str = {
    DHDR_TYPE_x86_pv  : "x86 PV",
    DHDR_TYPE_x86_hvm : "x86 HVM",
    DHDR_TYPE_x86_pvh : "x86 PVH",
    DHDR_TYPE_arm     : "ARM",
}

# Records
RH_FORMAT = "II"

REC_TYPE_end                  = 0x00000000
REC_TYPE_page_data            = 0x00000001
REC_TYPE_x86_pv_info          = 0x00000002
REC_TYPE_x86_pv_p2m_frames    = 0x00000003
REC_TYPE_x86_pv_vcpu_basic    = 0x00000004
REC_TYPE_x86_pv_vcpu_extended = 0x00000005
REC_TYPE_x86_pv_vcpu_xsave    = 0x00000006
REC_TYPE_shared_info          = 0x00000007
REC_TYPE_tsc_info             = 0x00000008
REC_TYPE_hvm_context          = 0x00000009
REC_TYPE_hvm_params           = 0x0000000a
REC_TYPE_toolstack            = 0x0000000b
REC_TYPE_x86_pv_vcpu_msrs     = 0x0000000c
REC_TYPE_verify               = 0x0000000d
REC_TYPE_checkpoint           = 0x0000000e

rec_type_to_str = {
    REC_TYPE_end                  : "End",
    REC_TYPE_page_data            : "Page data",
    REC_TYPE_x86_pv_info          : "x86 PV info",
    REC_TYPE_x86_pv_p2m_frames    : "x86 PV P2M frames",
    REC_TYPE_x86_pv_vcpu_basic    : "x86 PV vcpu basic",
    REC_TYPE_x86_pv_vcpu_extended : "x86 PV vcpu extended",
    REC_TYPE_x86_pv_vcpu_xsave    : "x86 PV vcpu xsave",
    REC_TYPE_shared_info          : "Shared info",
    REC_TYPE_tsc_info             : "TSC info",
    REC_TYPE_hvm_context          : "HVM context",
    REC_TYPE_hvm_params           : "HVM params",
    REC_TYPE_toolstack            : "Toolstack",
    REC_TYPE_x86_pv_vcpu_msrs     : "x86 PV vcpu msrs",
    REC_TYPE_verify               : "Verify",
    REC_TYPE_checkpoint           : "Checkpoint",
}

# page_data
PAGE_DATA_FORMAT             = "II"
PAGE_DATA_PFN_MASK           = (1L << 52) - 1
PAGE_DATA_PFN_RESZ_MASK      = ((1L << 60) - 1) & ~((1L << 52) - 1)

# flags from xen/public/domctl.h: XEN_DOMCTL_PFINFO_* shifted by 32 bits
PAGE_DATA_TYPE_SHIFT         = 60
PAGE_DATA_TYPE_LTABTYPE_MASK = (0x7L << PAGE_DATA_TYPE_SHIFT)
PAGE_DATA_TYPE_LTAB_MASK     = (0xfL << PAGE_DATA_TYPE_SHIFT)
PAGE_DATA_TYPE_LPINTAB       = (0x8L << PAGE_DATA_TYPE_SHIFT) # Pinned pagetable

PAGE_DATA_TYPE_NOTAB         = (0x0L << PAGE_DATA_TYPE_SHIFT) # Regular page
PAGE_DATA_TYPE_L1TAB         = (0x1L << PAGE_DATA_TYPE_SHIFT) # L1 pagetable
PAGE_DATA_TYPE_L2TAB         = (0x2L << PAGE_DATA_TYPE_SHIFT) # L2 pagetable
PAGE_DATA_TYPE_L3TAB         = (0x3L << PAGE_DATA_TYPE_SHIFT) # L3 pagetable
PAGE_DATA_TYPE_L4TAB         = (0x4L << PAGE_DATA_TYPE_SHIFT) # L4 pagetable
PAGE_DATA_TYPE_BROKEN        = (0xdL << PAGE_DATA_TYPE_SHIFT) # Broken
PAGE_DATA_TYPE_XALLOC        = (0xeL << PAGE_DATA_TYPE_SHIFT) # Allocate-only
PAGE_DATA_TYPE_XTAB          = (0xfL << PAGE_DATA_TYPE_SHIFT) # Invalid

# x86_pv_info
X86_PV_INFO_FORMAT        = "BBHI"

X86_PV_P2M_FRAMES_FORMAT  = "II"

# x86_pv_vcpu_{basic,extended,xsave,msrs}
X86_PV_VCPU_HDR_FORMAT    = "II"

# tsc_info
TSC_INFO_FORMAT           = "IIQII"

# hvm_params
HVM_PARAMS_ENTRY_FORMAT   = "QQ"
HVM_PARAMS_FORMAT         = "II"

class VerifyLibxc(VerifyBase):
    """ Verify a Libxc v2 stream """

    def __init__(self, info, read):
        VerifyBase.__init__(self, info, read)

        self.squashed_pagedata_records = 0


    def verify(self):
        """ Verity a libxc stream """

        self.verify_ihdr()
        self.verify_dhdr()

        while self.verify_record() != REC_TYPE_end:
            pass


    def verify_ihdr(self):
        """ Verify an Image Header """
        marker, ident, version, options, res1, res2 = \
            self.unpack_exact(IHDR_FORMAT)

        if marker != IHDR_MARKER:
            raise StreamError("Bad image marker: Expected 0x%x, got 0x%x"
                              % (IHDR_MARKER, marker))

        if ident != IHDR_IDENT:
            raise StreamError("Bad image id: Expected 0x%x, got 0x%x"
                              % (IHDR_IDENT, ident))

        if version != IHDR_VERSION:
            raise StreamError("Unknown image version: Expected %d, got %d"
                              % (IHDR_VERSION, version))

        if options & IHDR_OPT_RESZ_MASK:
            raise StreamError("Reserved bits set in image options field: 0x%x"
                              % (options & IHDR_OPT_RESZ_MASK))

        if res1 != 0 or res2 != 0:
            raise StreamError("Reserved bits set in image header: 0x%04x:0x%08x"
                              % (res1, res2))

        if ( (sys.byteorder == "little") and
             ((options & IHDR_OPT_BIT_ENDIAN) != IHDR_OPT_LE) ):
            raise StreamError(
                "Stream is not native endianess - unable to validate")

        endian = ["little", "big"][options & IHDR_OPT_LE]
        self.info("Libxc Image Header: %s endian" % (endian, ))


    def verify_dhdr(self):
        """ Verify a domain header """

        gtype, page_shift, res1, major, minor = \
            self.unpack_exact(DHDR_FORMAT)

        if gtype not in dhdr_type_to_str:
            raise StreamError("Unrecognised domain type 0x%x" % (gtype, ))

        if res1 != 0:
            raise StreamError("Reserved bits set in domain header 0x%04x"
                              % (res1, ))

        if page_shift != 12:
            raise StreamError("Page shift expected to be 12.  Got %d"
                              % (page_shift, ))

        if major == 0:
            self.info("Domain Header: legacy converted %s"
                      % (dhdr_type_to_str[gtype], ))
        else:
            self.info("Domain Header: %s from Xen %d.%d"
                      % (dhdr_type_to_str[gtype], major, minor))


    def verify_record(self):
        """ Verify an individual record """

        rtype, length = self.unpack_exact(RH_FORMAT)

        if rtype not in rec_type_to_str:
            raise StreamError("Unrecognised record type 0x%x" % (rtype, ))

        contentsz = (length + 7) & ~7
        content = self.rdexact(contentsz)

        if rtype != REC_TYPE_page_data:

            if self.squashed_pagedata_records > 0:
                self.info("Squashed %d Page Data records together"
                          % (self.squashed_pagedata_records, ))
                self.squashed_pagedata_records = 0

            self.info("Libxc Record: %s, length %d"
                      % (rec_type_to_str[rtype], length))

        else:
            self.squashed_pagedata_records += 1

        padding = content[length:]
        if padding != "\x00" * len(padding):
            raise StreamError("Padding containing non0 bytes found")

        if rtype not in record_verifiers:
            raise RuntimeError("No verification function for libxc record '%s'"
                               % rec_type_to_str[rtype])
        else:
            record_verifiers[rtype](self, content[:length])

        return rtype


    def verify_record_end(self, content):
        """ End record """

        if len(content) != 0:
            raise RecordError("End record with non-zero length")


    def verify_record_page_data(self, content):
        """ Page Data record """
        minsz = calcsize(PAGE_DATA_FORMAT)

        if len(content) <= minsz:
            raise RecordError("PAGE_DATA record must be at least %d bytes long"
                              % (minsz, ))

        count, res1 = unpack(PAGE_DATA_FORMAT, content[:minsz])

        if res1 != 0:
            raise StreamError("Reserved bits set in PAGE_DATA record 0x%04x"
                              % (res1, ))

        pfnsz = count * 8
        if (len(content) - minsz) < pfnsz:
            raise RecordError("PAGE_DATA record must contain a pfn record for "
                              "each count")

        pfns = list(unpack("=%dQ" % (count,), content[minsz:minsz + pfnsz]))

        nr_pages = 0
        for idx, pfn in enumerate(pfns):

            if pfn & PAGE_DATA_PFN_RESZ_MASK:
                raise RecordError("Reserved bits set in pfn[%d]: 0x%016x",
                                  idx, pfn & PAGE_DATA_PFN_RESZ_MASK)

            if pfn >> PAGE_DATA_TYPE_SHIFT in (5, 6, 7, 8):
                raise RecordError("Invalid type value in pfn[%d]: 0x%016x",
                                  idx, pfn & PAGE_DATA_TYPE_LTAB_MASK)

            # We expect page data for each normal page or pagetable
            if PAGE_DATA_TYPE_NOTAB <= (pfn & PAGE_DATA_TYPE_LTABTYPE_MASK) \
                    <= PAGE_DATA_TYPE_L4TAB:
                nr_pages += 1

        pagesz = nr_pages * 4096
        if len(content) != minsz + pfnsz + pagesz:
            raise RecordError("Expected %u + %u + %u, got %u"
                              % (minsz, pfnsz, pagesz, len(content)))


    def verify_record_x86_pv_info(self, content):
        """ x86 PV Info record """

        expectedsz = calcsize(X86_PV_INFO_FORMAT)
        if len(content) != expectedsz:
            raise RecordError("x86_pv_info: expected length of %d, got %d"
                              % (expectedsz, len(content)))

        width, levels, res1, res2 = unpack(X86_PV_INFO_FORMAT, content)

        if width not in (4, 8):
            raise RecordError("Expected width of 4 or 8, got %d" % (width, ))

        if levels not in (3, 4):
            raise RecordError("Expected levels of 3 or 4, got %d" % (levels, ))

        if res1 != 0 or res2 != 0:
            raise StreamError("Reserved bits set in X86_PV_INFO: 0x%04x 0x%08x"
                              % (res1, res2))

        bitness = {4:32, 8:64}[width]
        self.info("  %sbit guest, %d levels of pagetables" % (bitness, levels))


    def verify_record_x86_pv_p2m_frames(self, content):
        """ x86 PV p2m frames record """

        if len(content) % 8 != 0:
            raise RecordError("Length expected to be a multiple of 8, not %d"
                              % (len(content), ))

        start, end = unpack("=II", content[:8])
        self.info("  Start pfn 0x%x, End 0x%x" % (start, end))


    def verify_record_x86_pv_vcpu_generic(self, content, name):
        """ Generic for all REC_TYPE_x86_pv_vcpu_{basic,extended,xsave,msrs} """
        minsz = calcsize(X86_PV_VCPU_HDR_FORMAT)

        if len(content) <= minsz:
            raise RecordError("X86_PV_VCPU_%s record length must be at least %d"
                              " bytes long" % (name, minsz))

        vcpuid, res1 = unpack(X86_PV_VCPU_HDR_FORMAT, content[:minsz])

        if res1 != 0:
            raise StreamError(
                "Reserved bits set in x86_pv_vcpu_%s record 0x%04x"
                              % (name, res1))

        self.info("  vcpu%d %s context, %d bytes"
                  % (vcpuid, name, len(content) - minsz))


    def verify_record_shared_info(self, content):
        """ shared info record """

        if len(content) != 4096:
            raise RecordError("Length expected to be 4906 bytes, not %d"
                              % (len(content), ))


    def verify_record_tsc_info(self, content):
        """ tsc info record """

        sz = calcsize(TSC_INFO_FORMAT)

        if len(content) != sz:
            raise RecordError("Length should be %u bytes" % (sz, ))

        mode, khz, nsec, incarn, res1 = unpack(TSC_INFO_FORMAT, content)

        if res1 != 0:
            raise StreamError("Reserved bits set in TSC_INFO: 0x%08x"
                              % (res1, ))

        self.info("  Mode %u, %u kHz, %u ns, incarnation %d"
                  % (mode, khz, nsec, incarn))


    def verify_record_hvm_context(self, content):
        """ hvm context record """

        if len(content) == 0:
            raise RecordError("Zero length HVM context")


    def verify_record_hvm_params(self, content):
        """ hvm params record """

        sz = calcsize(HVM_PARAMS_FORMAT)

        if len(content) < sz:
            raise RecordError("Length should be at least %u bytes" % (sz, ))

        count, rsvd = unpack(HVM_PARAMS_FORMAT, content[:sz])

        if rsvd != 0:
            raise RecordError("Reserved field not zero (0x%04x)" % (rsvd, ))

        sz += count * calcsize(HVM_PARAMS_ENTRY_FORMAT)

        if len(content) != sz:
            raise RecordError("Length should be %u bytes" % (sz, ))


    def verify_record_toolstack(self, _):
        """ toolstack record """
        raise DeprecationWarning("Found Toolstack record in stream")


    def verify_record_verify(self, content):
        """ verify record """

        if len(content) != 0:
            raise RecordError("Verify record with non-zero length")


    def verify_record_checkpoint(self, content):
        """ checkpoint record """

        if len(content) != 0:
            raise RecordError("Checkpoint record with non-zero length")


record_verifiers = {
    REC_TYPE_end:
        VerifyLibxc.verify_record_end,
    REC_TYPE_page_data:
        VerifyLibxc.verify_record_page_data,

    REC_TYPE_x86_pv_info:
        VerifyLibxc.verify_record_x86_pv_info,
    REC_TYPE_x86_pv_p2m_frames:
        VerifyLibxc.verify_record_x86_pv_p2m_frames,

    REC_TYPE_x86_pv_vcpu_basic:
        lambda s, x:
        VerifyLibxc.verify_record_x86_pv_vcpu_generic(s, x, "basic"),
    REC_TYPE_x86_pv_vcpu_extended:
        lambda s, x:
        VerifyLibxc.verify_record_x86_pv_vcpu_generic(s, x, "extended"),
    REC_TYPE_x86_pv_vcpu_xsave:
        lambda s, x:
        VerifyLibxc.verify_record_x86_pv_vcpu_generic(s, x, "xsave"),
    REC_TYPE_x86_pv_vcpu_msrs:
        lambda s, x:
        VerifyLibxc.verify_record_x86_pv_vcpu_generic(s, x, "msrs"),

    REC_TYPE_shared_info:
        VerifyLibxc.verify_record_shared_info,
    REC_TYPE_tsc_info:
        VerifyLibxc.verify_record_tsc_info,

    REC_TYPE_hvm_context:
        VerifyLibxc.verify_record_hvm_context,
    REC_TYPE_hvm_params:
        VerifyLibxc.verify_record_hvm_params,
    REC_TYPE_toolstack:
        VerifyLibxc.verify_record_toolstack,
    REC_TYPE_verify:
        VerifyLibxc.verify_record_verify,
    REC_TYPE_checkpoint:
        VerifyLibxc.verify_record_checkpoint,
    }



#!/usr/bin/python

import sys,os

import idl

(TYPE_DEFBOOL, TYPE_BOOL, TYPE_INT, TYPE_UINT, TYPE_STRING, TYPE_ARRAY, TYPE_AGGREGATE) = range(7)

def py_type(ty):
    if ty == idl.bool:
        return TYPE_BOOL
    if ty.typename == "libxl_defbool":
        return TYPE_DEFBOOL
    if isinstance(ty, idl.Enumeration):
        return TYPE_UINT
    if isinstance(ty, idl.Number):
        if ty.signed:
            return TYPE_INT
        else:
            return TYPE_UINT
    if isinstance(ty, idl.Array):
        return TYPE_ARRAY
    if isinstance(ty, idl.Aggregate):
        return TYPE_AGGREGATE
    if ty == idl.string:
        return TYPE_STRING
    return None

def py_wrapstruct(ty):
    l = []
    l.append('typedef struct {')
    l.append('    PyObject_HEAD;')
    l.append('    %s obj;'%ty.typename);
    l.append('}Py_%s;'%ty.rawname)
    l.append('')
    return "\n".join(l) + "\n"

def fsanitize(name):
    "Sanitise a function name given a C type"
    ret = '_'.join(name.split())
    return ret.replace('*', 'ptr')

def py_decls(ty):
    l = []
    if isinstance(ty, idl.Aggregate):
        l.append('_hidden Py_%s *Py%s_New(void);\n'%(ty.rawname, ty.rawname))
        l.append('_hidden int Py%s_Check(PyObject *self);\n'%ty.rawname)
        for f in ty.fields:
            if py_type(f.type) is not None:
                continue
            if py_type(f.type) == TYPE_DEFBOOL:
                continue
            if ty.marshal_out():
                l.append('_hidden PyObject *attrib__%s_get(%s *%s);'%(\
                    fsanitize(f.type.typename), f.type.typename, f.name))
            if ty.marshal_in():
                l.append('_hidden int attrib__%s_set(PyObject *v, %s *%s);'%(\
                    fsanitize(f.type.typename), f.type.typename, f.name))
    return '\n'.join(l) + "\n"

def py_attrib_get(ty, f):
    t = py_type(f.type)
    l = []
    l.append('static PyObject *py_%s_%s_get(Py_%s *self, void *priv)'%(ty.rawname, f.name, ty.rawname))
    l.append('{')
    if t == TYPE_BOOL:
        l.append('    PyObject *ret;')
        l.append('    ret = (self->obj.%s) ? Py_True : Py_False;'%f.name)
        l.append('    Py_INCREF(ret);')
        l.append('    return ret;')
    elif t == TYPE_DEFBOOL:
        l.append('    return genwrap__defbool_get(&self->obj.%s);'%f.name)
    elif t == TYPE_INT:
        l.append('    return genwrap__ll_get(self->obj.%s);'%f.name)
    elif t == TYPE_UINT:
        l.append('    return genwrap__ull_get(self->obj.%s);'%f.name)
    elif t == TYPE_STRING:
        l.append('    return genwrap__string_get(&self->obj.%s);'%f.name)
    elif t == TYPE_AGGREGATE or t == TYPE_ARRAY:
        l.append('    PyErr_SetString(PyExc_NotImplementedError, "Getting %s");'%ty.typename)
        l.append('    return NULL;')
    else:
        tn = f.type.typename
        l.append('    return attrib__%s_get((%s *)&self->obj.%s);'%(fsanitize(tn), tn, f.name))
    l.append('}')
    return '\n'.join(l) + "\n\n"

def py_attrib_set(ty, f):
    t = py_type(f.type)
    l = []
    l.append('static int py_%s_%s_set(Py_%s *self, PyObject *v, void *priv)'%(ty.rawname, f.name, ty.rawname))
    l.append('{')
    if t == TYPE_BOOL:
        l.append('    self->obj.%s = (NULL == v || Py_None == v || Py_False == v) ? 0 : 1;'%f.name)
        l.append('    return 0;')
    elif t == TYPE_DEFBOOL:
        l.append('    return genwrap__defbool_set(v, &self->obj.%s);'%f.name)
    elif t == TYPE_UINT or t == TYPE_INT:
        l.append('    %slong long tmp;'%(t == TYPE_UINT and 'unsigned ' or ''))
        l.append('    int ret;')
        if t == TYPE_UINT:
            l.append('    ret = genwrap__ull_set(v, &tmp, (%s)~0);'%f.type.typename)
        else:
            l.append('    ret = genwrap__ll_set(v, &tmp, (%s)~0);'%f.type.typename)
        l.append('    if ( ret >= 0 )')
        l.append('        self->obj.%s = tmp;'%f.name)
        l.append('    return ret;')
    elif t == TYPE_STRING:
        l.append('    return genwrap__string_set(v, &self->obj.%s);'%f.name)
    elif t == TYPE_AGGREGATE or t == TYPE_ARRAY:
        l.append('    PyErr_SetString(PyExc_NotImplementedError, "Setting %s");'%ty.typename)
        l.append('    return -1;')
    else:
        tn = f.type.typename
        l.append('    return attrib__%s_set(v, (%s *)&self->obj.%s);'%(fsanitize(tn), tn, f.name))
    l.append('}')
    return '\n'.join(l) + "\n\n"

def py_object_def(ty):
    l = []
    if ty.dispose_fn is not None:
        dtor = '    %s(&self->obj);\n'%ty.dispose_fn
    else:
        dtor = ''

    funcs="""static void Py%(rawname)s_dealloc(Py_%(rawname)s *self)
{
%(dtor)s    self->ob_type->tp_free((PyObject *)self);
}

static int Py%(rawname)s_init(Py_%(rawname)s *self, PyObject *args, PyObject *kwds)
{
    memset(&self->obj, 0, sizeof(self->obj));
    return genwrap__obj_init((PyObject *)self, args, kwds);
}

static PyObject *Py%(rawname)s_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    Py_%(rawname)s *self = (Py_%(rawname)s *)type->tp_alloc(type, 0);
    if (self == NULL)
        return NULL;
    memset(&self->obj, 0, sizeof(self->obj));
    return (PyObject *)self;
}

"""%{'rawname': ty.rawname, 'dtor': dtor}

    l.append('static PyGetSetDef Py%s_getset[] = {'%ty.rawname)
    for f in ty.fields:
        if f.type.private:
            continue
        l.append('    { .name = "%s", '%f.name)
        if ty.marshal_out():
            l.append('      .get = (getter)py_%s_%s_get, '%(ty.rawname, f.name))
        else:
            l.append('      .get = (getter)NULL, ')
        if ty.marshal_in():
            l.append('      .set = (setter)py_%s_%s_set,'%(ty.rawname, f.name))
        else:
            l.append('      .set = (setter)NULL,')
        l.append('    },')
    l.append('    { .name = NULL }')
    l.append('};')
    struct="""
static PyTypeObject Py%s_Type= {
    PyObject_HEAD_INIT(NULL)
    0,
    PKG ".%s",
    sizeof(Py_%s),
    0,
    (destructor)Py%s_dealloc,     /* tp_dealloc        */
    NULL,                         /* tp_print          */
    NULL,                         /* tp_getattr        */
    NULL,                         /* tp_setattr        */
    NULL,                         /* tp_compare        */
    NULL,                         /* tp_repr           */
    NULL,                         /* tp_as_number      */
    NULL,                         /* tp_as_sequence    */
    NULL,                         /* tp_as_mapping     */
    NULL,                         /* tp_hash           */
    NULL,                         /* tp_call           */
    NULL,                         /* tp_str            */
    NULL,                         /* tp_getattro       */
    NULL,                         /* tp_setattro       */
    NULL,                         /* tp_as_buffer      */
    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE, /* tp_flags          */
    "%s",                         /* tp_doc            */
    NULL,                         /* tp_traverse       */
    NULL,                         /* tp_clear          */
    NULL,                         /* tp_richcompare    */
    0,                            /* tp_weaklistoffset */
    NULL,                         /* tp_iter           */
    NULL,                         /* tp_iternext       */
    NULL,                         /* tp_methods        */
    NULL,                         /* tp_members        */
    Py%s_getset,                  /* tp_getset         */
    NULL,                         /* tp_base           */
    NULL,                         /* tp_dict           */
    NULL,                         /* tp_descr_get      */
    NULL,                         /* tp_descr_set      */
    0,                            /* tp_dictoffset     */
    (initproc)Py%s_init,          /* tp_init           */
    NULL,                         /* tp_alloc          */
    Py%s_new,                     /* tp_new            */
};

Py_%s *Py%s_New(void)
{
    return (Py_%s *)Py%s_new(&Py%s_Type, NULL, NULL);
}

int Py%s_Check(PyObject *self)
{
    return (self->ob_type == &Py%s_Type);
}
"""%tuple(ty.rawname for x in range(15))
    return funcs + '\n'.join(l) + "\n" + struct

def py_initfuncs(types):
    l = []
    l.append('void genwrap__init(PyObject *m)')
    l.append('{')
    for ty in types:
        if isinstance(ty, idl.Enumeration):
            for v in ty.values:
                l.append('    PyModule_AddIntConstant(m, "%s", %s);' % (v.rawname, v.name))
        elif isinstance(ty, idl.Aggregate):
            l.append('    if (PyType_Ready(&Py%s_Type) >= 0) {'%ty.rawname)
            l.append('        Py_INCREF(&Py%s_Type);'%ty.rawname)
            l.append('        PyModule_AddObject(m, "%s", (PyObject *)&Py%s_Type);'%(ty.rawname, ty.rawname))
            l.append('    }')
        else:
            raise NotImplementedError("unknown type %s (%s)" % (ty.typename, type(ty)))        

    l.append('}')
    return '\n'.join(l) + "\n\n"

def tree_frob(types):
    ret = types[:]
    for ty in [ty for ty in ret if isinstance(ty, idl.Aggregate)]:
        ty.fields = filter(lambda f:f.name is not None and f.type.typename is not None, ty.fields)
    return ret

if __name__ == '__main__':
    if len(sys.argv) < 4:
        print >>sys.stderr, "Usage: genwrap.py <idl> <decls> <defns>"
        sys.exit(1)

    (_,types) = idl.parse(sys.argv[1])

    types = tree_frob(types)

    decls = sys.argv[2]
    f = open(decls, 'w')
    f.write("""#ifndef __PYXL_TYPES_H
#define __PYXL_TYPES_H

/*
 * DO NOT EDIT.
 *
 * This file is autogenerated by
 * "%s"
 */

#define PKG "xen.lowlevel.xl"

#if __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 1)
#define _hidden __attribute__((visibility("hidden")))
#define _protected __attribute__((visibility("protected")))
#else
#define _hidden
#define _protected
#endif

/* Initialise all types */
_hidden void genwrap__init(PyObject *m);

/* Generic type initialiser */
_hidden int genwrap__obj_init(PyObject *self, PyObject *args, PyObject *kwds);

/* Auto-generated get/set functions for simple data-types */
_hidden int genwrap__string_set(PyObject *v, char **str);
_hidden PyObject *genwrap__string_get(char **str);
_hidden PyObject *genwrap__ull_get(unsigned long long val);
_hidden int genwrap__ull_set(PyObject *v, unsigned long long *val, unsigned long long mask);
_hidden PyObject *genwrap__ll_get(long long val);
_hidden int genwrap__ll_set(PyObject *v, long long *val, long long mask);
_hidden PyObject *genwrap__defbool_get(libxl_defbool *db);
_hidden int genwrap__defbool_set(PyObject *v, libxl_defbool *db);

""" % " ".join(sys.argv))
    for ty in [ty for ty in types if isinstance(ty, idl.Aggregate)]:
        f.write('/* Internal API for %s wrapper */\n'%ty.typename)
        f.write(py_wrapstruct(ty))
        f.write(py_decls(ty))
        f.write('\n')
    f.write('#endif /* __PYXL_TYPES_H */\n')
    f.close()
 
    defns = sys.argv[3]
    f = open(defns, 'w')
    f.write("""/*
 * DO NOT EDIT.
 *
 * This file is autogenerated by
 * "%s"
 */

#include <Python.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include "libxl.h" /* gah */
#include "%s"

""" % tuple((' '.join(sys.argv),) + (os.path.split(decls)[-1:]),))
    for ty in types:
        if ty.private:
            continue
        if isinstance(ty, idl.Aggregate):
            f.write('/* Attribute get/set functions for %s */\n'%ty.typename)
            for a in ty.fields:
                if a.type.private:
                    continue
                if ty.marshal_out():
                    f.write(py_attrib_get(ty,a))
                if ty.marshal_in():
                    f.write(py_attrib_set(ty,a))
            f.write(py_object_def(ty))
    f.write(py_initfuncs(types))
    f.close()



#!/usr/bin/python

import sys;
from structs import structs, compat_arches;

# command line arguments
outfile = sys.argv[1];
archs   = sys.argv[2:];

f = open(outfile, "w");
f.write('''
/*
 * sanity checks for generated foreign headers:
 *  - verify struct sizes
 *
 * generated by %s -- DO NOT EDIT
 */
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <inttypes.h>
''');

for a in archs:
    f.write('#include "%s.h"\n' % a);

f.write('int main(int argc, char *argv[])\n{\n');

f.write('\tprintf("\\n");');
f.write('printf("%-25s |", "structs");\n');
for a in archs:
    f.write('\tprintf("%%8s", "%s");\n' % a);
f.write('\tprintf("\\n");');

f.write('\tprintf("\\n");');
for struct in structs:
    f.write('\tprintf("%%-25s |", "%s");\n' % struct);
    for a in archs:
        s = struct + "_" + a;
        if compat_arches.has_key(a):
            compat = compat_arches[a]
            c = struct + "_" + compat;
        else:
            compat = None
        f.write('#ifdef %s_has_no_%s\n' % (a, struct));
        f.write('\tprintf("%8s",\n');
        if compat:
            f.write('# ifndef %s_has_no_%s\n' % (compat, struct));
            f.write('\t\t"!"\n');
            f.write('# else\n')
            f.write('\t\t"-"\n');
            f.write('# endif\n')
        else:
            f.write('\t\t"-"\n');
        f.write('\t);\n')
        f.write("#else\n");
        f.write('\tprintf("%%8zd", sizeof(struct %s));\n' % s);
        if compat:
            f.write('\tif (sizeof(struct %s) != sizeof(struct %s))\n' % (s, c))
            f.write('\t\tprintf("!");\n')
        f.write("#endif\n");

    f.write('\tprintf("\\n");\n\n');

f.write('\tprintf("\\n");\n');
f.write('\texit(0);\n');
f.write('}\n');

f.close();

